import dotenv from 'dotenv'
import { setConfig, getConfig, getLogger, generateEvent } from 'protobase';
import { getBaseConfig, getConfigWithoutSecrets } from '@my/config'
import { pathToFileURL } from 'url';

// get config vars
dotenv.config({ path: '../../.env' });
global.defaultRoute = '/api/core/v1'
global.appName = 'core'
import { getServiceToken, getApp, getMQTTClient } from 'protonode'
setConfig(getBaseConfig("core", process, getServiceToken()))
import adminModules from './api'
require('events').EventEmitter.defaultMaxListeners = 100;
const { createExpressProxy, handleUpgrade } = require('app/proxy.js')

import http from 'http';

import chokidar from 'chokidar';
import BundleContext from 'app/bundles/coreContext'
import { startMqtt } from './mqtt';

export const startCore = (ready?) => {
  const config = getConfig()
  const logger = getLogger()

  const app = getApp((app) => app.use( createExpressProxy('core') ))

  process.on('uncaughtException', function (err) {
    logger.error({ err }, 'Uncaught Exception: ', err.message)
  });

  startMqtt(config)
  logger.info({ config: getConfigWithoutSecrets(config) }, "Service Started: core")
  logger.debug({ adminModules }, 'Admin modules: ', JSON.stringify(adminModules))
  const mqtt = getMQTTClient('core', getServiceToken())

  const topicSub = (mqtt, topic, cb) => {
    mqtt.subscribe(topic)
    mqtt.on("message", (messageTopic, message) => {
      const isWildcard = topic.endsWith("#");
      if (!isWildcard && topic != messageTopic) {
        return
      }
      if (isWildcard && !messageTopic.startsWith(topic.slice(0, -1).replace(/\/$/, ''))) {
        return
      }
      const parsedMessage = message.toString();
      cb(parsedMessage, messageTopic)
    });
  };

  const topicPub = (mqtt, topic, data) => {
    mqtt.publish(topic, data)
  }

  try {
    
    import(pathToFileURL(require.resolve('app/bundles/coreApis')).href).then((BundleAPI) => {
      BundleAPI.default(app, { mqtt, topicSub, topicPub, ...BundleContext })
    })

  } catch (error) {
    logger.error({ error: error.toString() }, "Server error")
  }

  const server = http.createServer(app);
  handleUpgrade(server, 'core')
  const PORT = 8000
  server.listen(PORT, () => {
    logger.debug({ service: { protocol: "http", port: PORT } }, "Service started: HTTP")
    if(ready) {
      ready(PORT, getServiceToken())
    }
    if (process.send) {
      //notify potential fork parents about the service readiness
      process.send('ready');
    } else {
      //if there is no fork, generate a start event
      generateEvent({
        path: 'services/core/start', //event type: / separated event category: files/create/file, files/create/dir, devices/device/online
        from: 'core', // system entity where the event was generated (next, api, cmd...)
        user: 'system', // the original user that generates the action, 'system' if the event originated in the system itself
        payload: {}, // event payload, event-specific data
      }, getServiceToken())
    }
  });

  const isFullDev = process.env.FULL_DEV === '1';

  if (isFullDev) {
    const pathsToWatch = [
      'src/**',
      '../../extensions/**',
      '../../packages/app/conf.ts',
      '../../packages/protolib/dist/**',
      '../../packages/app/bundles/coreApis.ts',
      '../../packages/app/bundles/coreContext.ts',
      '../../packages/protonode/dist/**',
      '../../packages/protobase/dist/**',
    ];

    const watcher = chokidar.watch(pathsToWatch, {
      ignored: /^([.][^.\/\\])|([\/\\]+[.][^.])/,
      persistent: true
    });

    var restarting = false
    var restartTimer = null
    watcher.on('change', async (path) => {
      if (restarting) {
        clearTimeout(restartTimer)
      } else {
        console.log(`File ${path} has been changed, restarting...`);
        restarting = true
      }

      restartTimer = setTimeout(async () => {
        await generateEvent({
          path: 'services/core/stop', //event type: / separated event category: files/create/file, files/create/dir, devices/device/online
          from: 'core', // system entity where the event was generated (next, api, cmd...)
          user: 'system', // the original user that generates the action, 'system' if the event originated in the system itself
          payload: {}, // event payload, event-specific data
        }, getServiceToken())
        process.exit(0)
      }, 1000);
    })
  }
}

if (require.main === module) {
  startCore()
}
