"use strict";
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.ProtoSqliteDB = void 0;
const fs_1 = __importDefault(require("fs"));
const path_1 = __importDefault(require("path"));
const better_sqlite3_1 = __importDefault(require("better-sqlite3"));
const protodb_1 = require("../protodb");
const protobase_1 = require("protobase");
const logger = (0, protobase_1.getLogger)();
class ProtoSqliteDB extends protodb_1.ProtoDB {
    constructor(dbPath, options, config) {
        super(dbPath, options, config);
        this.capabilities = ['pagination', 'groupBySingle', 'groupByOptions'];
        this.db = new better_sqlite3_1.default(dbPath, { fileMustExist: false, ...options });
        this.prepareSchema();
    }
    /** Crea las tablas si no existen */
    prepareSchema() {
        this.db.exec(`
      CREATE TABLE IF NOT EXISTS entries (
        id          TEXT PRIMARY KEY,
        data        TEXT NOT NULL,
        created_at  INTEGER NOT NULL
      );
      CREATE TABLE IF NOT EXISTS indexTable (
        name  TEXT PRIMARY KEY,
        value TEXT NOT NULL
      );
    `);
    }
    static connect(dbPath, options, config) {
        return new ProtoSqliteDB(dbPath, options, config);
    }
    static getInstance(dbPath, options, config) {
        return new ProtoSqliteDB(dbPath, options, config);
    }
    /** Inicializa en disco y precarga initialData */
    static async initDB(dbPath, initialData = {}, options) {
        const dir = path_1.default.dirname(dbPath);
        if (!fs_1.default.existsSync(dir))
            fs_1.default.mkdirSync(dir, { recursive: true });
        const db = new better_sqlite3_1.default(dbPath, { fileMustExist: false });
        db.exec(`
      CREATE TABLE IF NOT EXISTS entries (
        id          TEXT PRIMARY KEY,
        data        TEXT NOT NULL,
        created_at  INTEGER NOT NULL
      );
      CREATE TABLE IF NOT EXISTS indexTable (
        name  TEXT PRIMARY KEY,
        value TEXT NOT NULL
      );
    `);
        if (options === null || options === void 0 ? void 0 : options.indexes) {
            db.prepare(`INSERT OR REPLACE INTO indexTable(name, value) VALUES('indexes', ?)`).run(JSON.stringify(options.indexes));
        }
        if (options === null || options === void 0 ? void 0 : options.groupIndexes) {
            db.prepare(`INSERT OR REPLACE INTO indexTable(name, value) VALUES('groupIndexes', ?)`).run(JSON.stringify(options.groupIndexes));
        }
        const now = Date.now();
        const insert = db.prepare(`INSERT OR IGNORE INTO entries(id, data, created_at) VALUES(?,?,?)`);
        for (const [id, obj] of Object.entries(initialData)) {
            insert.run(id, JSON.stringify(obj), now);
        }
        db.close();
    }
    // — CRUD básico —
    get(key) {
        const row = this.db
            .prepare(`SELECT data FROM entries WHERE id = ?`)
            .get(key);
        if (!row)
            throw new Error('NotFound');
        return row.data;
    }
    async exists(key) {
        const row = this.db
            .prepare(`SELECT 1 FROM entries WHERE id = ?`)
            .get(key);
        return !!row;
    }
    async put(key, value, options) {
        const now = Date.now();
        const existing = this.db
            .prepare(`SELECT created_at FROM entries WHERE id = ?`)
            .get(key);
        const created_at = existing ? existing.created_at : now;
        this.db
            .prepare(`INSERT INTO entries(id,data,created_at)
         VALUES(@id,@data,@created_at)
         ON CONFLICT(id) DO UPDATE SET data=@data`)
            .run({ id: key, data: value, created_at });
        return true;
    }
    async del(key) {
        this.db.prepare(`DELETE FROM entries WHERE id = ?`).run(key);
        return true;
    }
    // — Conteos e índices —
    async count(filter) {
        if (filter) {
            const sql = `
        SELECT COUNT(*) AS c FROM entries
        WHERE json_extract(data, ?) = ?
      `;
            // utilizamos placeholder para la ruta JSON
            return this.db
                .prepare(sql)
                .get(`$.${filter.key}`, filter.value).c;
        }
        return this.db.prepare(`SELECT COUNT(*) AS c FROM entries`).get().c;
    }
    async getIndexedKeys() {
        try {
            const row = this.db
                .prepare(`SELECT value FROM indexTable WHERE name='indexes'`)
                .get();
            return JSON.parse(row.value).keys;
        }
        catch {
            return [];
        }
    }
    async getGroupIndexes() {
        try {
            const row = this.db
                .prepare(`SELECT value FROM indexTable WHERE name='groupIndexes'`)
                .get();
            return JSON.parse(row.value);
        }
        catch {
            return [];
        }
    }
    async hasGroupIndexes(keys) {
        const gis = await this.getGroupIndexes();
        return keys.every(k => gis.some((gi) => gi.key === k));
    }
    async getGroupIndexOptions(groupKey, limit = 100) {
        const sql = `
      SELECT DISTINCT json_extract(data, ?) AS v
      FROM entries
      WHERE v IS NOT NULL
      LIMIT ?
    `;
        // la ruta JSON va como parámetro
        return this.db
            .prepare(sql)
            .all(`$.${groupKey}`, limit)
            .map((r) => r.v);
    }
    // — Paginación ordenada —
    async getPageItems(total, key, pageNumber, itemsPerPage, direction, filter) {
        // validamos la dirección
        const dir = direction.toLowerCase() === 'desc' ? 'DESC' : 'ASC';
        let sql = `SELECT data FROM entries`;
        const params = [];
        if (filter) {
            sql += ` WHERE json_extract(data, ?) = ?`;
            params.push(`$.${filter.key}`, filter.value);
        }
        sql += `
      ORDER BY json_extract(data, ?) ${dir}
      LIMIT ? OFFSET ?
    `;
        params.push(`$.${key}`, itemsPerPage, pageNumber * itemsPerPage);
        return this.db
            .prepare(sql)
            .all(...params)
            .map((r) => r.data);
    }
    // — Iterador para list() —
    *iterator() {
        const stmt = this.db.prepare(`SELECT id, data FROM entries`);
        for (const row of stmt.iterate()) {
            yield [row.id, row.data];
        }
    }
    async close() {
        this.db.close();
    }
}
exports.ProtoSqliteDB = ProtoSqliteDB;
